/*jslint bitwise: true, eqeqeq: true, nomen: true, undef: true */
/*global jQuery, console, window, document, Option */

(function($) {
/**
 * jquery.mvc.js
 * A simple model controller for jQuery. (ie. client side MVC)
 *
 * Copyright (c) 2008 Mark Gibson
 * Dual licensed under the MIT (MIT-LICENSE.txt)
 * and GPL (GPL-LICENSE.txt) licenses.
 *
 * There is a single model which is simply a Javascript object, stored in
 * $.mvc.model
 * it is intended to contain more objects/arrays in a hierarchical fashion,
 * so that values may be referenced with a simple path string.
 * Example: "/user/3/info/name"
 *
 * The model should be capable of being serialized to JSON, therefore it should
 * not contain functions, or circular references.
 *
 * $.mvc(ref) produces an object which represents the path to the data in the
 *    model. Operations may be chained to this as in jQuery, example:
 *
 * $.mvc("/user/name").val('Mark').bind(function()...)
 *
 * The ref may be a string path, eg: "/user/3/info/name", or an array path, eg:
 * ['user',3,'info','name'], or as DOM element (from which a model reference
 * will be obtained through the 'ref' attribute hierarchy).
 *
 * The MVC object may also be obtained from the first element of a jQuery set,
 * eg: $('#blah').mvc()
 *
 * @author Mark Gibson
 */
var version='0.14';

// Private functions & var shortcuts, defined later
var init, parseRef, opts, binds, Trail, Path, SingleValue, MultiValue, log;

// Utility functions

// The identity function
function idfn(v) { return v; }

// No-operation function
function nofn() {}

// Set 'checked' property of an element. If the element has a value attribute
// then the argument must match that, otherwise the strings 'true', 'false'
// or the boolean value of the argument determines the checked state.
function setChecked(v) {
	if (v === undefined) { return; }
	var node = this.getAttributeNode('value');
	if (node) {
		// Set as 'checked' only if the arg matches the 'value' attr
		this.checked = (v === node.nodeValue);
	} else {
		this.checked = (v === 'true') || (v === 'false' ? false : !!v);
	}
	return this.checked;
}

// Get the value of a 'checked' element.
function getChecked() {
	if (this.disabled) { return; }
	var node = this.getAttributeNode('value');
	if (node) {
		// A 'value' attr is present, so return the value only if checked
		return this.checked ? node.nodeValue : undefined;
	}
	// Return the checked status if the 'value' attr is not present
	return this.checked;
}

// Main MVC constructor function
$.mvc = function () {
	if (arguments.length === 0) {
		var trail = new Trail();
		trail.step = trail.length = 0;
		return trail;
	}
	return init.apply(arguments[0], arguments);
};

// MVC namespace
$.extend($.mvc, {
	// The global data model.
	model: {},

	// Global options that affect the model functions
	options: {
		debug: false,

		/**
		 * Create an array when a non-existent parent contains a numeric segment name.
		 * Example: for the path: '/user/3', if the /user doesn't exist it will be created as an array
		 * if this option is true, otherwise it is created as an object.
		 */
		createArrays: true,

		/**
		 * Disable automatic setup of repeated elements, to allow manual setup
		 * of more complex needs.
		 * FIXME: Repeated elements need to be analysed for dependencies so that
		 * a manual setup is not required.
		 */
		noRepeat: false,

		// Selector for elements which can be updated by model changes
		initSelector: 'label,span,option,:input,.mvc,[mvc-type]',

		// Selector for elements that don't generate 'change' events
		noChange: ($.browser.msie ? ':button,:checkbox,:radio,[mvc-type=radiogroup]' : ':button'),

		// Regular Expressions
		reTest: /[\{:]/,
		reModel: /\{([^\s\{\}:;]+)\}/,
		reName: /^([\w\d\_\-\.\@])+$/,
		reExpr: /(\:)([\w\d\_\-]+)(\(([^\)]*)\))?/,
		reFilter: /\[([^\]]+)\]/,
//		reCompare: /([\w\d\_\-\.])+((=)([^\]]*))?/,
		reRoot: /^(\/|\:repeat\-path)/
	},

	// List of elements or functions bound to model values.
	bindings: {},

	// Default handlers for 'change' and 'update' events when an element is bound to the model
	handler: {
		change: function(event) {
			log('event', 'change', this, event);
			// Update the model from the element value
			$.mvc(this).val($.mvc.queryElement(this), event.target);
		},
		update: function(event, modelEvent) {
			log('event', 'update', this, event, modelEvent);
			event.stopPropagation();
			$.mvc.updateElement(this, modelEvent);
		},
		triggerChange: function() {
			log('event', 'triggerChange', this);
			$(this).trigger('change');
		}
	},

	// Methods for setting the value of an element
	setter: {
		'default': function(v) { $(this).val(v); },
		none: nofn,
		checkbox: setChecked,
		radio: setChecked,
		button: nofn,
		reset: nofn,
		submit: nofn,
		plain: function(v) { $(this).text(v.toString()); },
		html: function(v) { $(this).html(v.toString()); },
		radiogroup: function(v) {
			$(':radio', this).each(function() {
				this.checked = (this.value === v);
			});
		},
		para: function(v) {		// Split a string into paragraphs
			$($.trim(v).split('\n'))
				.map(function(n,c) { return $('<p>').text(c)[0]; })
				.appendTo($(this).empty());
		},
		csv: function(v) {
			$(this).text(v && v.join ? v.join(', ') : v);
		}
	},

	// Methods for getting the value from an element
	getter: {
		'default': function() { return this.disabled ? undefined : $(this).val(); },
		checkbox: getChecked,
		radio: getChecked,
		radiogroup: function() {
			return $(':checked',this).val();
		}
	}
});

if ($.browser.msie) {
	// In IE, the value prop of a button contains the inner text of the button,
	// so we add a getter here to get the value attribute instead.
	// (BTW, getAttribute('value') also returns the inner content!!!)
	$.mvc.getter.button = function() {
		return this.getAttributeNode('value').value;
	};
}

opts = $.mvc.options; // shortcut to options
binds = $.mvc.bindings;

log = function() {
	// Allow logging of only certain types of message
	if (opts.debug && (opts.debug === true || opts.debug.indexOf(arguments[0]) >= 0)) {
		console.log.apply(console, arguments);
	}
};

init = function(ref, refAttr) {
	var path, node, relative,
		deps = [];	// dependencies of this ref

	if (!refAttr) { refAttr = 'ref'; }

	// Is ref a DOMElement?
	if (ref && ref.nodeType) {
		node = ref;
		var refs = [];

		var accumRefs = function() {
			var attr = this.getAttribute(refAttr);
			if (attr !== null) {
				attr = $.mvc.sub(attr, node, deps, refAttr, true);
				if (attr === undefined) {
					// An undefined substitution
					ref = undefined;
					return false;
				}
				refs.unshift(attr);
				// Have we reached the model root?
				if (opts.reRoot.test(attr)) { return false; }
			}
		};

		if (accumRefs.call(ref) !== false) {
			$(node).parents('['+refAttr+']').each(accumRefs);
		}

		if (ref !== undefined) {
			ref = refs.join('/');
		}
	}

	if (typeof ref === 'string') {

		ref = $.mvc.sub(ref, node, deps, refAttr, true);

		if (ref !== undefined) {
			ref = ref.split('/');
			if (ref[0] === '') {
				// An absolute path
				ref.shift();
			} else {
	//			console.warn('Attempt to use a relative path in mvc!');
				relative = true;
			}
		}
	}

	// Is ref Array like
	if (ref && typeof ref === 'object' && typeof ref.length === 'number' && !ref.nodeType) {

		ref = $.makeArray(ref);

		$.each(ref, function(n, expr) {
			if (!opts.reName.test(expr)) {
				path = new Path();
				return false;
			}
		});

		if (!path) { path = new Trail(); }

		if (relative) { path.relative = true; }
		if (deps.length) { path.deps = deps; }
		if (node) { path.element = node; }

		path.length = 0;
		Array.prototype.push.apply(path, ref);

		log('init', path);

		return path.toLeaf();
	}

	path = new Trail();
	path.invalid = true;

	log('init', 'invalid ref', path);

	return path;
};

// ---- Trail - simple single-valued reference --------------------------------

$.extend($.mvc, {
	/**
	 * Substitute model values into a template string.
	 *
	 * Model values placeholders are represented by a reference between curly braces.
	 * Example:
	 * "Page {/pager/page} of {/pager/pages}"
	 *
	 */
	sub: function(template, element, deps, attr, strict) {

		if (opts.reTest.test(template)) {
			var m, val;
			var tmp = template; // DEBUG ONLY

			log('sub', 'strict', strict);

			while ( (m = opts.reExpr.exec(template)) !== null ) {
				val = $.mvc.expr[m[1]][m[2]].call(element, m[4], deps, attr);
				if (strict && val === undefined) { return; }

				tmp = template;
				template = template.replace(m[0], val);

				log('sub', 'expr', tmp, ' -> ', template);
			}

			while ( (m = opts.reModel.exec(template)) !== null ) {
				// FIXME: Should also bind to the model and
				// perform update on changes rather than here

				val = $.mvc(m[1]).val();
				if (strict && val === undefined) { return; }

				tmp = template;
				template = template.replace(m[0], val);

				log('sub', 'lookup', tmp, ' -> ', template);
			}
		}
		return template;
	},

	expr: {
		':': {
			'repeat-key': function(p, deps, attr) {
				var repeat = $(this)[attr === 'mvc-repeat' ? 'parents' : 'selfOrParents']('[mvc-repeat]');
				if (repeat.length) {
					if (deps) { deps.push(repeat.get(0)); }
					return repeat.mvc('mvc-repeat').key();
				}
			},
			'repeat-path': function(p, deps, attr) {
				var repeat = $(this)[attr === 'mvc-repeat' ? 'parents' : 'selfOrParents']('[mvc-repeat]');
				if (repeat.length) {
					if (deps) { deps.push(repeat.get(0)); }
					return repeat.attr('mvc-repeat');
				}
			}
		}
	},

	walk: {
		// An outgoing function that constructs the required objects/arrays
		// as it walks.
		construct: function(parent, key, path) {
			// Construct object/arrays as we walk the path
			if (parent[key] === undefined) {
				parent[key] = (isNaN(path[this.length]) || !opts.createArrays) ? {} : [];
			}
		},

		// An incoming function that triggers model events
		trigger: function(event) {
			this.trigger(event);
		}
	},

	// Update the elements from the model

	updateElement: function(elem, modelEvent) {
		// Don't update an element if it was the source of an event chain
		if (modelEvent && modelEvent.source === elem) { return; }

		var value = (modelEvent && modelEvent.value !== undefined) ?
					modelEvent.value : $.mvc(elem).val();

		// TODO: Maybe extend this to allow a filter chain
		var filterAttr = elem.getAttribute('mvc-out');
		if (filterAttr === 'key') {
			value = $.mvc(elem).key();
		}

		// Model value is undefined
		if (value === undefined) { return; }

		// Determine the function used to render the element value
		var fn = $.data(elem, 'mvc-setter') ||
				 $.mvc.setter[elem.getAttribute('mvc-type') || elem.type || 'plain'] ||
				 $.mvc.setter['default'];

		log('view', 'updateElement', elem, value);

		if (fn) {
			fn.call(elem, value);
		}

		return value;
	},

	// Get a value from an element.

	queryElement: function(elem) {
		// Determine the function used to read the element value
		var fn = $.data(elem, 'mvc-getter') ||
				 $.mvc.getter[elem.getAttribute('mvc-type') || elem.type || 'plain'] ||
				 $.mvc.getter['default'];

		var value = fn ? fn.call(elem) : undefined;

		log('view', 'queryElement', elem, value);

		return value;
	}

});

Trail = function () {};

$.mvc.trail = Trail.prototype = {
	mvc: version,
	type: 'trail',

	toString: function() {
		return (this.relative ? '' : '/') +
			Array.prototype.slice.call(this, 0, this.step+1).join('/');
	},

	// Walk the path, applying functions upto the current step

	walk: function(params) {
		var target = this.step,
			parent = $.mvc.model,
			key,
			reply;

		log('walk', params);

		this.pushStep();

		for (this.step = 0; this.step < target; this.step++) {
			key = this[this.step];

			log('walk', 'outgoing', this.step, this, parent, key, parent[key]);

			// Call the outgoing callback function
			if (params.outgoing) {
				params.outgoing.call(this, parent, key, this);
			}

			// Have we reached an abrupt end?
			if (parent[key] === undefined) { break; }

			parent = parent[key];
		}

		if (params.target && this.step === target) {
			key = this[target];

			log('walk', 'target', this.step, this, parent, key);

			// Call the target callback function
			reply = params.target.call(this, parent, key);
		}

		if (reply && params.incoming) {
			do {
				log('walk', 'incoming', this.step, this, reply);

				// Call the incoming callback function
				params.incoming.call(this, reply);

				this.step--;
			} while (this.step >= 0);
		}

		return this.popStep();
	},

	valueObject: function() {
		return new SingleValue();
	},

	// Get/Set a value (or values) at the current step along the path

	val: function() {
		var value, source;

		// Are we being called to GET a value
		if (arguments.length === 0) {

			value = this.valueObject();

			// Walk the path upto the current step
			this.walk({
				target: function(parent, key) { value.set(parent[key], this); }
			});

			log('model', 'get', this, ' = ', value);

			return value.valueOf();

		} else {

			// We should SET the value
			value = arguments[0];
			source = arguments[1];

			// Don't set a value to undefined
			if (value === undefined) { return; }

			return this.walk({
				target: function(parent, key) {
					var type = key in parent ? 'set' : 'add',
						was = parent[key];
					
					// If value is a function, then call it to retreive the value
					if ($.isFunction(value)) {
						value = value.call(this, was, parent, key);
					}

					if (value !== was) {
						// Change the value
						parent[key] = value;

						log('model', type, this, ' = ', value);

						// Reply with an event
						return {
							type:   type,
							target: $.mvc(this),
							value:  value,
							was:    was,
							source: source
						};
					}
				},
				outgoing: $.mvc.walk.construct,
				incoming: $.mvc.walk.trigger
			});
		}
	},

	// Delete the value at the current step.

	del: function(source) {

		// Walk the path upto the current step
		return this.walk({
			target: function(parent, key) {
				var was = parent[key];
				if (was !== undefined) {

//					if (parent.splice) {
						// Remove the value from the array
//						parent.splice(key, 1);

						// FIXME: Using splice may not be safe, as it will alter
						// the key of all following items in the array.
						// May be better to set the value to undefined, and
						// normalize the array after the walk.

//					} else {
						// Remove the property from the object
						delete parent[key];
//					}

					log('model', 'delete', this);

					// Reply with an event
					return {
						type:   'delete',
						target: $.mvc(this),
						was:    was,
						source: source
					};
				}
			},
			incoming: $.mvc.walk.trigger
		});
	},

	// Get/Set the key at the current step

	key: function(expr) {
		if (arguments.length === 0) {
			return this[this.step];
		}
		// TODO: Test expr is simple, if not convert to a path
		this[this.step] = expr;
		return this;
	},

	// Add one or more steps to the end of the path, and move to the end

	addStep: function() {
		Array.prototype.push.apply(this, arguments);
		return this.toLeaf();
	},

	// Remove the last step from the path, and move to the end

	removeLastStep: function() {
		Array.prototype.pop.apply(this);
		return this.toLeaf();
	},

	/**
	 * Trigger a model event.
	 *
	 * The event param can be just the event type string,
	 * or an event object containing the properties:
	 *   type   - event type
	 *   target - the target of the event (absolute ref, no expressions)
	 *   value  - the current, or new value (optional)
	 *   was    - the old value (optional)
	 *   source - the element which changed the value, this will not be updated (optional)
	 *
	 * or an Array of these objects
	 */
	trigger: function(event) {
		var trail = this, ref = this.toString(), bindings = binds[ref];

		if (bindings) {
			if (typeof event === 'string') {
				event = {
					type: event,
					target: this
				};
			}

			log('event', 'trigger', this, event);

			$.each(bindings, function(n, bound) {
				bound(trail, event);
			});
		}
		return this;
	},

	bind: function(bound) {
		var ref = this.toString();

		log('binding', 'bind', ref, ' -> ', bound);

		if (binds[ref] === undefined) {
			binds[ref] = [];
		}

		var bindings = binds[ref];

		function triggerEach (fn, type) {
			return function(trail, event) {

				log('event', 'triggerEach', trail, event);

				if (!event.length) {
					if (type && type.indexOf(event.type) < 0) { return; }
					// Call the fn for just a single event
					fn.call(trail, event);
				} else {
					// Call the bound function for every event
					$.each(event, function(n, event) {
						if (type && type.indexOf(event.type) < 0) { return; }
						fn.call(trail, event);
					});
				}
			};
		}
		function triggerBatch (fn) {
			return function(trail, event) {
				if (!event.length) { event = [event]; }

				log('event', 'triggerEach', trail, event);

				// Call the function passing a batch of events
				fn.call(trail, event);
			};
		}
		function triggerElem (elem, type) {
			return function(trail, event) {
				// Use only the very last event if given an array
				if (event.length) { event = event[event.length-1]; }

				if (type && type.indexOf(event.type) < 0) { return; }

				log('event', 'triggerElem', trail, event);

				if (elem !== event.source) {
					// Trigger an 'update' event on the element
					$(elem).trigger('update', [event, trail]);
				}
			};
		}

		if ($.isFunction(bound)) {
			// The bound function will be called for every event
			bindings.push(triggerEach(bound));

		} else if (bound.nodeType) {
			bindings.push(triggerElem(bound));

		} else if (typeof bound === 'object') {
			if ($.isFunction(bound.batch)) {
				bindings.push(triggerBatch(bound.batch));
			}
			if ($.isFunction(bound.each)) {
				bindings.push(triggerEach(bound.each, bound.type));
			}
			if ($.isFunction(bound.batchEnd)) {
				bindings.push(triggerBatch(bound.batchEnd));
			}
		}

		return this;
	},

	unbind: function() {
		log('binding', 'unbind', this);

		delete binds[this.toString()];
		return this;
	},

	// Move to the given step along the path

	toStep: function(step) {
		this.step = step;
		return this;
	},

	// Move to the end (leaf) of the path

	toLeaf: function() {
		this.step = this.length-1;
		return this;
	},

	// Move up the path, toward the root.
	// The argument 'n' is the number of steps to jump, default: 1.

	up: function(n) {
		if (n === undefined) { n = 1; }
		this.step -= n;
		if (this.step < 0) { this.step = 0; }
		return this;
	},

	// Set defaults for non-existent values

	def: function(defaults) {
		return this.val($.extend({}, defaults, this.val()));
	},

	// Push the current step onto a stack

	pushStep: function() {
		if (!this.stepStack) { this.stepStack = []; }
		this.stepStack.push(this.step);
		return this;
	},

	// Move to the step last pushed onto the stack

	popStep: function() {
		this.step = this.stepStack.pop();
		return this;
	},

	// Return back to the jQuery set

	jQuery: function() {
		return this.jQuerySaved;
	}
};

SingleValue = function () {};

SingleValue.prototype = {
	set: function(value) {
		this.value = value;
		return this;
	},

	valueOf: function() {
		return this.value;
	},

	toString: function() {
		return this.value === undefined ? undefined : this.value.toString();
	}
};


// ---- Path - an expressive multi-valued reference ---------------------------

$.extend($.mvc, {

	copyAttrs: function(src, dst) {
		// Copy attributes from the place holder into the clone
		$.each(src.attributes, function(n, a) {
			dst.setAttribute(a.nodeName, a.nodeValue);
		});
	},

	// Setup a repeated element

	repeat: function(elems) {
		var cache = {}, repeated = [];

		elems.each(function() {
			var template = $(this),
				path = template.mvc('mvc-repeat'),
				pathLen = path.length,
				parent = this.parentNode,
				selector = template.attr('mvc-template');

			log('repeat', 'start', this, path, parent);

			// Abort if the path is invalid
			if (path.invalid) {
				log('repeat', 'invalid path');
				return;
			}

			if (path.type === 'trail') {
				// No need to repeat a trail, it has only one value
				repeated.push(this);
				log('repeat', 'trail');
				return;
			}

			// Should we get the repeat template from elsewhere?
			if (selector) {
				log('repeat', 'template from', selector);

				// Find and clone the template
				template = template.parents(selector).eq(0).clone();

				// Copy attributes from the placeholder to the template
				$.mvc.copyAttrs(this, template.get(0));

				// Remove the template attribute
				template.removeAttr('mvc-template');
			}

			// Remove the repeat node from the document
			$(this).remove();

			// Save the template element
			$.data(parent, opts.mvcData+'-repeat', template.get(0));


			var add = function(ref) {
				var clone = template.clone().attr('mvc-repeat', ref);

				// HACK ALERT: IE doesn't take kindly to messing with it's SELECT elements
				if ($.browser.msie && parent.nodeName === 'SELECT' && clone.get(0).nodeName === 'OPTION') {
					(function() {
						// Create an Option object
						var option = new Option();

						// Copy attributes to the OPTION
						$.mvc.copyAttrs(clone.get(0), option);

						// Add the OPTION to the SELECT, in a way befitting to IE
						parent.options[parent.options.length] = option;
					})();

					// Return the option element
					return $('> option:last', parent).get(0);
				}

				log('repeat', 'add', clone[0]);

				// Append the clone to the parent element
				return clone.appendTo(parent).get(0);
			};

			path.toLeaf().walk({
				target: function(p, key) {
					var ref = this.toString(),
						clone = add(ref);

					cache[ref] = clone;
					repeated.push(clone);
				}
			})
			.up()
			.bind({type: 'add',
				each: function(ev) {
					// Ensure we only add a new element if the event occurs
					// at the right path step.
					if (ev.target.length === pathLen) {
						var ref = ev.target.toString(),
							clone = add(ref);
						cache[ref] = clone;
						$(clone).mvcInit();
						$(clone).parent().trigger('mvc-repeat-add', clone);
					}
				}
			})
			.bind({type: 'delete',
				each: function(ev) {
					var ref = ev.target.toString();
					if (cache[ref] && ev.target.length === pathLen) {
						$(cache[ref])
							.parent()
								.trigger('mvc-repeat-delete', cache[ref])
							.end()
							.remove();
						delete cache[ref];
					}
				}
			});
		});

		log('repeat', 'repeated', repeated);

		$(repeated).each(function() {
			var node = this;

			// Find only the shallowest repeated descendents
			var next =
				$(this).find('[mvc-repeat]')
					.filter(function() {
						return ($(this).parents('[mvc-repeat]:first').get(0) === node);
					});

			log('repeat', 'next', next.get());

			if (next.length > 0) {
				$.mvc.repeat(next);
			}
		});
	},

	// Expand a step expression into an array of keys

	expandStepExpr: function(parent, expr) {
		var keys = [];

		if (expr === '*') {
			// A wildcard expression, find all keys
			$.each(parent, function(key) { keys.push(key); });
		} else if (typeof expr === 'object' && typeof expr.length === 'number') {
			// An array of key names
			keys = expr;

		} else if (typeof expr === 'string') {
			if (opts.reName.test(expr)) {
				// A single key name
				keys.push(expr);
			} else {
				// Parse a filter
				var m = opts.reFilter.exec(expr);
				if (m) {
					$.each(m[1].split(','), function(n, key) {
						// TODO: Check for more complex expressions here
						keys.push(key);
					});
					log('parse', 'filter', keys);
				}
			}
		}

		return keys;
	}
});

Path = function () {};

$.mvc.path = Path.prototype = new Trail(false);

$.extend($.mvc.path, {
	type: 'path',

	// Walk the path

	walk: function(params) {
		var	trail = $.mvc(),			// Breadcrumb trail
			path = this,
			target = this.step;

		params = $.extend({
			target: nofn,
			outgoing: nofn,
			incoming: nofn
		}, params);

		log('walk', this);

		function explore(parent) {
			var step = trail.length,	// The current step is always the length of our trail
				expr = path[step],		// The step expression
				reply,
				replies = [],			// Replies from the target callbacks
				allow;					// Set to false if an item is to be filtered out

			log('walk', 'explore', trail, parent);

			// Expand the expression to an array of keys, and iterate
			$.each($.mvc.expandStepExpr(parent, expr), function(n, key) {

				var filtered = false;

				// Apply any filters registered at this step
				if (path.filters && path.filters[step]) {
					$.each(path.filters[step], function() {
						return (filtered = !this(parent[key], parent, key));
					});
				}

				log('walk', 'key', key, filtered);

				if (filtered) { return; }

				// Add to the breadcrumb trail
				trail.addStep(key);

				// Have we reached our target?
				if (step === target) {

					log('walk', 'target', trail, parent, key);

					// Call the target callback function
					reply = params.target.call(trail, parent, key);
					if (reply !== undefined) {
						replies.push(reply);
					}

				} else {

					log('walk', 'outgoing', trail, parent, key, path);

					// Call the outgoing callback function
					params.outgoing.call(trail, parent, key, path);

					// Explore the next step, if it exists
					if (parent[key] !== undefined) {
						replies = replies.concat(explore(parent[key]));
					}
				}

				if (replies.length) {
					log('walk', 'incoming', trail, replies);

					// Call the incoming callback function
					params.incoming.call(trail, replies);
				}

				// Pick up the breadcrumbs as we retreat
				trail.removeLastStep();
			});

			// Take the replies back along the trail
			return replies;
		}

		explore($.mvc.model); // Start at the root of the model

		return this;
	},

	valueObject: function() {
		return new MultiValue();
	},

	/**
	 * Bind a function or element to a model property.
	 *
	 * When the property is modified the function is called
	 * or the 'update' event is triggered on the element.
	 *
	 * The callback function will get an absolute path string as 'this',
	 * and an event object as the first argument, see $.mvc.trigger() function
	 * for more details.
	 */
	bind: function(bound) {
		return this.walk({
			target: function() {
				$.mvc(this.toString()).bind(bound);
			},
			outgoing: $.mvc.walk.construct
		});
	},

	// Unbind all elements/functions from the model value.

	unbind: function() {
		return this.walk({
			target: function() {
				$.mvc(this.toString()).unbind();
			}
		});
	},

	// Trigger a model update event, upto the current step.

	trigger: function(event) {
		if (typeof event === 'string') {
			event = { type: event };
		}

		return this.walk({
			target: function(parent, key) {
				event.target = $.mvc(this);
				return event;
			},
			incoming: $.mvc.walk.trigger
		});
	},

	// Register a filter

	filter: function(fn) {
		if (!this.filters) { this.filters = []; }
		if (!this.filters[this.step]) { this.filters[this.step] = []; }

		this.filters[this.step].push(fn);

		return this;
	}

});

MultiValue = function () {};

MultiValue.prototype = {
	set: function(value, ref) {
		log('value', 'set', ref.toString(), '=', value);
		this['$'+ref.toString()] = value;
		return this;
	},

	joinWith: ' ',

	valueOf: function() {
		return this;
	},

	keys: function() {
		var a = [];
		for(var k in this) {
			if (k.charAt(0) === '$') {
				a.push(k.substring(1));
			}
		}
		return a;
	},

	get: function(k) {
		return this['$'+k];
	},

	toArray: function() {
		var a = [];
		for(var k in this) {
			if (k.charAt(0) === '$') {
				a.push(this[k]);
			}
		}
		return a;
	},

	toString: function() {
		return this.toArray().join(this.joinWith);
	}
};

// ---- jQuery extensions -----------------------------------------------------

$.fn.extend({

	// Convert the jQuery set into an MVC path.

	mvc: function(refAttr) {
		var mvc = $.mvc(this[0], refAttr);
		mvc.jQuerySaved = this;
		return mvc;
	},

	// Initialise the elements and all descendents.
	// Returns the jQuery set of all bound elements.

	mvcInit: function() {

		// Deal with the repeated elements first
		if (!$.mvc.options.noRepeat) {
			$.mvc.repeat(
				// Reverse the set, so we setup the deeper repeats first
				Array.prototype.reverse.apply(
					// Find all repeating elements, and filter out ones with dependencies
					$('[mvc-repeat]', this).filter(function() {
						return !$.mvc(this, 'mvc-repeat').deps;
					})
				)
			);
		}

		return this
			.find('[mvc-sub]')
				.mvcSub()					// Setup template substitutions
			.end()
			.find(opts.initSelector)
				.mvcBind()
					.filter(':not(:button)')
						.each(function() {
							if ($.mvc.updateElement(this) === undefined) {
								// Model value is undefined for this element
								// so trigger the element to set the value
								$(this).triggerHandler('change');
							}
						})
					.end();
	},

	mvcSub: function() {
		return this.each(function() {
			var that = $(this);
			$.each(that.attr('mvc-sub').split(' '), function(i, attr) {
				that.attr(attr, $.mvc.sub(that.attr(attr), that[0]));
			});
		});
	},

	/**
	 * Bind the model to the elements, so that changes to the element values
	 * are reflected automatically in the model and vis versa.
	 *
	 * If model data exists for the element, then the value of the element
	 * is set from this data.
	 *
	 * Returns the set of bound elements, use .end() to revert to the original set.
	 */
	mvcBind: function() {

		return this
			.filter('[ref]')
				.each(function() {
					$.mvc(this).bind(this);
				})
				.bind('change.mvc', $.mvc.handler.change)
				.bind('update.mvc', $.mvc.handler.update)
				.filter(opts.noChange)
					.bind('click.mvc', $.mvc.handler.change)
				.end();
	},

	/**
	 * Unbind the jQuery elements from the model.
	 */
	mvcUnbind: function() {
		return this
			.filter('[ref]')
				.each(function() {
					$.mvc(this).unbind();
				})
				.unbind('change.mvc')
				.unbind('update.mvc')
				.unbind('click.mvc')
			.end();
	}
});

// jQuery extensions:
if (!$.fn.selfOrParents) {
	$.fn.selfOrParents = function( selector ) {
		var ret = jQuery.map( this, function(elem) {
			var matched = [], cur = elem;
			while ( cur && cur !== document ) {
				if ( cur.nodeType === 1 ) {
					matched.push( cur );
				}
				cur = cur.parentNode;
			}
			return matched;
		});

		if ( selector && typeof selector === "string" ) {
			ret = jQuery.multiFilter( selector, ret );
		}
		return this.pushStack( jQuery.unique( ret ) );
	};
}

})(jQuery);

